"""Module containing functions to round data to the standard rounding needed
for display.

Contains:
- round_thousands_to_1dp: Rounds values to 1dp in steps of 1000.
"""


def round_thousands_to_1dp(value: float) -> float:
    """
    Rounds values to 1dp in steps of 1000.
    Rounds values greater than 1B to nearest 100M
    Rounds values > 1B and >= 1M to nearest 100k
    Rounds values > 1M and >= 1k to nearest 100
    Rounds values less than 1000 to 1dp.
    If value is not a number, return original value.

    Args:
        value (float): Value to round

    Returns:
        float: Rounded value
    """
    try:
        if value >= 1_000_000_000 or value <= -1_000_000_000:
            return round(value, -8)
        if value >= 1_000_000 or value <= -1_000_000:
            return round(value, -5)
        if value >= 1000 or value <= -1_000:
            return round(value, -2)
        return round(value, 1)

    except TypeError:
        return value



"""Classes to be used for defined enums and type structures"""

from enum import Enum
from typing import TypedDict


class XAxisFormat(Enum):
    """Enum for date format on x axis"""

    YEAR = "year"
    MONTH_YEAR = "month_year"
    MONTH_YEAR_MONTHLY_DATA = "month_year_monthly_data"
    FINANCIAL_QUARTER = "financial_quarter"


class TitleDataStructure(TypedDict):
    """A TypedDict representing the structure of title_data"""

    MAIN_TITLE: str
    SUBTITLE: str


class HoverDataStructure(TypedDict):
    """A TypedDict representing the structure of hover_data"""

    CUSTOM_DATA: list[str]
    HOVER_TEXT_HEADERS: list[str]


class HoverDataByTrace(TypedDict):
    """A TypedDict representing hover_data organized by tracename"""

    tracename: dict[
        str, HoverDataStructure
    ]  # Each tracename maps to a HoverDataStructure



"""test_barchart_data"""

import polars as pl
from gov_uk_dashboards.components.plotly.stacked_barchart import StackedBarChart


def validate_barchart_instance(barchart_instance: StackedBarChart, **context):
    """
    Helper function to validate a barchart_instance dataframe.

    Args:
        barchart_instance: The barchart object with dataframe and metadata
        **context: Arbitrary key-value pairs (e.g., la="LA1", expenditure_area="Health").
            These will be included in the assertion message if provided.
    """
    barchart_df = barchart_instance.df
    trace_name_list = barchart_instance.trace_name_list
    trace_name_column = barchart_instance.trace_name_column
    x_axis_column = barchart_instance.x_axis_column
    unique_x_axis_values = set(barchart_df[x_axis_column].unique())
    context_str = ", ".join(f"{k}={v}" for k, v in context.items())

    for trace_name in trace_name_list:
        if not trace_name_column:
            df = barchart_df
        else:
            trace_values = barchart_df[trace_name_column].unique().to_list()
            if trace_name not in trace_values:
                raise AssertionError(
                    f"Trace '{trace_name}' expected in column '{trace_name_column}' "
                    f"but not found for {context_str}"
                )
            df = barchart_df.filter(pl.col(trace_name_column) == trace_name)

        # Check trace has at least one row
        assert df.height > 0, f"Trace '{trace_name}' has no data for {context_str}"

        # Check trace has all required x-axis values
        x_values_for_trace = set(df[x_axis_column].unique())
        assert x_values_for_trace == unique_x_axis_values, (
            f"{unique_x_axis_values - x_values_for_trace} missing for trace '{trace_name}' for "
            f"{context_str}"
        )

        # Check x-axis uniqueness
        assert (
            df[x_axis_column].is_unique().to_numpy()[0]
        ), f"x_axis-column: {x_axis_column} contains duplicate values. For {context_str}"



"""http_headers"""

import os
import dash


def setup_application_http_response_headers(dash_app: dash.Dash):
    """Add HTTP headers to the response"""
    server = dash_app.server

    @server.after_request
    def add_headers(response):
        content_security_policy = (
            "default-src 'self' 'unsafe-eval' 'unsafe-inline' data:; "
            "script-src 'self' 'unsafe-inline' https://*.googletagmanager.com "
            "https://cdn.jsdelivr.net/npm/web-vitals@4.2.4/dist/web-vitals.min.js "
            "https://cdn.jsdelivr.net/npm/html2canvas@1.4.1/dist/html2canvas.min.js "
            "https://*.google-analytics.com;"
            "connect-src 'self' https://*.googletagmanager.com https://*.google-analytics.com "
            "https://*.analytics.google.com; "
            "img-src 'self' https://*.googletagmanager.com https://*.google-analytics.com "
            "https://*.analytics.google.com data: https://*.tile.openstreetmap.org; "
        )
        frame_ancestors = os.environ.get("ALLOWED_FRAME_ANCESTORS")
        if frame_ancestors:
            content_security_policy += "; frame-ancestors " + frame_ancestors
        response.headers.add("Content-Security-Policy", content_security_policy)

        response.headers.add("X-Content-Type-Options", "nosniff")
        response.headers.add("Referrer-Policy", "no-referrer")
        response.headers.add("X-Frame-Options", "DENY")

        # The below header will disable all browser features for the dashboard website
        response.headers.add(
            "Permission-Policy",
            "accelerometer=(), ambient-light-sensor=(), autoplay=(), battery=(), camera=(), "
            "cross-origin-isolated=(), display-capture=(), document-domain=(), encrypted-media=()"
            ", execution-while-not-rendered=(), execution-while-out-of-viewport=(), fullscreen=(),"
            " geolocation=(), gyroscope=(), keyboard-map=(), magnetometer=(), microphone=(), "
            "midi=(), navigation-override=(), payment=(), picture-in-picture=(), "
            "publickey-credentials-get=(), screen-wake-lock=(), sync-xhr=(), usb=(), web-share=()"
            ", xr-spatial-tracking=()",
        )
        return response



"""tooltip"""

from dash import html
import dash_bootstrap_components as dbc


def tooltip(
    tooltip_text: str, tooltip_id: str, color: str = "white", font_size: str = "16px"
):
    """Creates a tooltip for providing further information

    Args:
        tooltip_text (str): The text to display when hovering
        tooltip_id (str): The element ID to contain the tooltip text
        color (str): The color of the tooltip text
        font_size (str): The size of the tooltip text

    Returns:
        html.Span: The dash HTML object for the tooltip.
    """

    return html.Span(
        [
            " ⓘ",
            dbc.Tooltip(
                tooltip_text,
                target=tooltip_id,
                class_name="tooltip tooltiptext",
                style={"color": color, "fontSize": font_size},
            ),
        ],
        id=tooltip_id,
        role="tooltip",
        **{"aria-label": tooltip_text},
        style={"cursor": "pointer"},
    )



"""Module containing enums used in the construction of plotly figures.

Contains:
- DashPatterns: Sets out valid dash patterns used by plotly/plotly express.
"""

from .dash_patterns import DashPatterns



# Import required functions
from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="UTF-8")

# Call setup function
setup(
    author="Department for Levelling Up, Housing and Communities",
    description="Provides access to functionality common to creating a data dashboard.",
    name="gov_uk_dashboards",
    version="25.6.1",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "setuptools~=59.8",
        "dash~=3.0",
        "numpy>=2.3.2",
        "dash_bootstrap_components~=2.0.3",
        "plotly~=6.2.0",
        "flask-basicauth~=0.2.0",
    ],
    include_package_data=True,
)



"""Module containing tools to assist with formatting of numbers in dashboards.

Contains:
- format_as_human_readable: Format a number as a human readable string, with
    options to customise elements such as prefix and suffix.
- rounding: Functions to round data to the standard rounding we want for display.
    - round_thousands_to_1dp: Rounds values to 1dp in steps of 1000.
"""

from .human_readable import format_as_human_readable
from . import rounding



"""Functions to add formatting to number values"""


def format_number_into_thousands_or_millions(
    number: int, thousand_decimal_places: int = 0
) -> str:
    """Format number into thousands or millions, eg. 1,500 becomes 1.5k & 1,234,567 becomes 1.235m

    Args:
        number (int): Integer to format.
        thousand_decimal_places (int or str): The number of decimal places to display for
            1_000<=number<1_000_000. Defaults to 0. If "default" is passed, number is simply
            divided by 1_000.
    """
    if number >= 1_000_000:
        formatted_number = f"{number / 1_000_000:.3f}m"
    elif number >= 1_000:
        if thousand_decimal_places == "default":
            formatted_number = f"{number / 1_000}k"
        else:
            formatted_number = f"{number / 1_000:.{thousand_decimal_places}f}k"
    else:
        formatted_number = str(number)
    return formatted_number



"""Functions to format text"""

import re


def create_id_from_string(string):
    """Function to create an id from a string. Remove non alphanumeric characters and replaces
    spaces with dashes"""
    if string is None:
        return ""
    return re.sub(r"[^a-z0-9]+", "-", string.lower()).strip("-")
